/*********************************************************************
*                      SEGGER Microcontroller GmbH                   *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*      (c) 2018  SEGGER Microcontroller GmbH                         *
*                                                                    *
*       www.segger.com     Support: support@segger.com               *
*                                                                    *
**********************************************************************
-------------------------- END-OF-HEADER -----------------------------

File        : MainTask_IOT_HTTP_GetWeatherData.c
Purpose     : Issue a GET request over a plain socket.

*/

/*********************************************************************
*
*       #include Section
*
**********************************************************************
*/

#include "RTOS.h"
#include "BSP.h"
#include "IP.h"
#include "IOT.h"
#include "WeatherStation.h"
#include <stdlib.h>

/*********************************************************************
*
*       Defines
*
**********************************************************************
*/

#define USE_RX_TASK  0  // 0: Packets are read in ISR, 1: Packets are read in a task of its own.

/*********************************************************************
*
*       enum
*
**********************************************************************
*/
enum {
   TASK_PRIO_IP_APPLICATION = 150
  ,TASK_PRIO_IP_TASK          // Priority must be higher as all IP application tasks.
#if USE_RX_TASK
  ,TASK_PRIO_IP_RX_TASK       // Must be the highest priority of all IP related tasks.
#endif
};

/*********************************************************************
*
*       Local types
*
**********************************************************************
*/

typedef struct {
  int Socket;
} CONNECTION_CONTEXT;

typedef enum {
  JSON_TYPE_OBJECT,
  JSON_TYPE_ARRAY,
  JSON_TYPE_PAIR,
  JSON_TYPE_STRING,
  JSON_TYPE_NUMBER,
  JSON_TYPE_TRUE,
  JSON_TYPE_FALSE,
  JSON_TYPE_NULL,
} JSON_TYPE;

typedef struct JSON_NODE JSON_NODE;
struct JSON_NODE {
  JSON_TYPE   Type;    // Type of this node
  JSON_NODE * pNext;   // Next node in list (linking arrays and object keys)
  JSON_NODE * pParent; // Pointer to node's parent
  JSON_NODE * pValue;  // Pointer to value (JSON_TYPE_KEY, JSON_TYPE_ARRAY, JSON_TYPE_OBJECT)
  char      * sStr;    // Only valid for JSON_TYPE_PAIR and JSON_TYPE_STRING
  float      Number;   // Only valid for JSON_TYPE_NUMBER
};

typedef struct {
  void* pKey;
  int WeatherNow;
  int WeatherDayIndex;
} USER_CONTEXT;

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/

static IP_HOOK_ON_STATE_CHANGE _StateChangeHook;
static int                     _IFaceId;
static OS_TIMER                _TIMERPulse;

//
// Task stacks and Task-Control-Blocks.
//
static OS_STACKPTR int _IPStack[TASK_STACK_SIZE_IP_TASK/sizeof(int)];               // Stack of the IP_Task.
static OS_TASK         _IPTCB;                                                      // Task-Control-Block of the IP_Task.

#if USE_RX_TASK
static OS_STACKPTR int _IPRxStack[TASK_STACK_SIZE_IP_RX_TASK/sizeof(int)];          // Stack of the IP_RxTask.
static OS_TASK         _IPRxTCB;                                                    // Task-Control-Block of the IP_RxTask.
#endif

static WEATHER_STATUS _aWeatherDay[NUM_FORECASTS_PER_DAY * NUM_DAYS_FORECAST];

static char           _aBuf[2048];              // Only used in _GetWeather(). Declared as static to avoid task stack usage.
static char           _aPayload[2048];          // Only used in _GetWeather(). Declared as static to avoid task stack usage.


/*********************************************************************
*
*       Prototypes
*
**********************************************************************
*/

static void _BeginObject(IOT_JSON_CONTEXT *pContext);
static void _EndObject  (IOT_JSON_CONTEXT *pContext);
static void _BeginArray (IOT_JSON_CONTEXT *pContext);
static void _EndArray   (IOT_JSON_CONTEXT *pContext);
static void _Key        (IOT_JSON_CONTEXT *pContext, const char *sText);
static void _String     (IOT_JSON_CONTEXT *pContext, const char *sText);
static void _Number     (IOT_JSON_CONTEXT *pContext, const char *sText);
static void _Literal    (IOT_JSON_CONTEXT *pContext, const char *sText);

/*********************************************************************
*
*       Static const data
*
**********************************************************************
*/

static const IOT_JSON_EVENTS _EventAPI = {
  _BeginObject,
  _EndObject,
  _BeginArray,
  _EndArray,
  _Key,
  _String,
  _Number,
  _Literal
};


/*********************************************************************
*
*       Static code
*
**********************************************************************
*/

/*********************************************************************
*
*       _BeginObject()
*
*  Function description
*    Handle Begin Object event.
*
*  Parameters
*    pContext - Pointer to JSON context.
*/
static void _BeginObject(IOT_JSON_CONTEXT *pContext) {
  (void)pContext;
}

/*********************************************************************
*
*       _EndObject()
*
*  Function description
*    Handle End Object event.
*
*  Parameters
*    pContext - Pointer to JSON context.
*/
static void _EndObject(IOT_JSON_CONTEXT *pContext) {
  (void)pContext;
}

/*********************************************************************
*
*       _BeginArray()
*
*  Function description
*    Handle Begin Array event.
*
*  Parameters
*    pContext - Pointer to JSON context.
*/
static void _BeginArray(IOT_JSON_CONTEXT *pContext) {
  (void)pContext;
}

/*********************************************************************
*
*       _EndArray()
*
*  Function description
*    Handle End Array event.
*
*  Parameters
*    pContext - Pointer to JSON context.
*/
static void _EndArray(IOT_JSON_CONTEXT *pContext) {
  (void)pContext;
}

/*********************************************************************
*
*       _Key()
*
*  Function description
*    Handle Key event.
*
*  Parameters
*    pContext - Pointer to JSON context.
*    sText    - Zero-terminated key ID.
*/
static void _Key(IOT_JSON_CONTEXT *pContext, const char *sText) {
  USER_CONTEXT*   pUserContext;
  WEATHER_STATUS* pStatus;

  pUserContext = IOT_JSON_GetUserContext(pContext);
  if(pUserContext->WeatherNow) {
    pStatus = &MAIN_WeatherNow;
  } else {
    pStatus = &_aWeatherDay[pUserContext->WeatherDayIndex];
  }
  if (strcmp(sText, "temp") == 0) {
    pUserContext->pKey = pStatus->acTemp;
  } else if (strcmp(sText, "temp_min") == 0) {
    pUserContext->pKey = pStatus->acTempMin;
  } else if (strcmp(sText, "temp_max") == 0) {
    pUserContext->pKey = pStatus->acTempMax;
  } else if (strcmp(sText, "main") == 0) {
    pUserContext->pKey = pStatus->acWeather;
  } else if (strcmp(sText, "icon") == 0) {
    pUserContext->pKey = pStatus->acWeatherIcon;
  } else if (strcmp(sText, "humidity") == 0) {
    pUserContext->pKey = pStatus->acHumidity;
  } else if (strcmp(sText, "dt") == 0) {
    pUserContext->pKey = pStatus->acWeekday;
  } else if(strcmp(sText, "dt_txt") == 0 && !pUserContext->WeatherNow) { // last key for the day
    pUserContext->pKey = pStatus->acDate;
    pUserContext->WeatherDayIndex++;
  } else {
    pUserContext->pKey = NULL;
  }
}

/*********************************************************************
*
*       _String()
*
*  Function description
*    Handle Key event.
*
*  Parameters
*    pContext - Pointer to JSON context.
*    sText    - Zero-terminated string value.
*/
static void _String(IOT_JSON_CONTEXT *pContext, const char* sText) {
  USER_CONTEXT* pUserContext;
  //
  pUserContext = IOT_JSON_GetUserContext(pContext);
  if (pUserContext->pKey != NULL) {
    strcpy(pUserContext->pKey, sText);
    pUserContext->pKey = NULL;
  }
}

/*********************************************************************
*
*       _Number()
*
*  Function description
*    Handle Number event.
*
*  Parameters
*    pContext - Pointer to JSON context.
*    sText    - Zero-terminated numeric value.
*/
static void _Number(IOT_JSON_CONTEXT* pContext, const char* sText) {
  USER_CONTEXT* pUserContext;
  //
  pUserContext = IOT_JSON_GetUserContext(pContext);
  if (pUserContext->pKey != NULL) {
    strcpy(pUserContext->pKey, sText);
    pUserContext->pKey = NULL;
  }
}

/*********************************************************************
*
*       _Literal()
*
*  Function description
*    Handle Literal event.
*
*  Parameters
*    pContext - Pointer to JSON context.
*    sText    - Zero-terminated literal name.
*/
static void _Literal(IOT_JSON_CONTEXT* pContext, const char* sText) {
  USER_CONTEXT* pUserContext;
  //
  pUserContext = IOT_JSON_GetUserContext(pContext);
  if (pUserContext->pKey != NULL) {
    strcpy(pUserContext->pKey, sText);
    pUserContext->pKey = NULL;
  }
}

/*********************************************************************
*
*       _OnStateChange()
*
* Function description
*   Callback that will be notified once the state of an interface
*   changes.
*
* Parameters
*   IFaceId   : Zero-based interface index.
*   AdminState: Is this interface enabled ?
*   HWState   : Is this interface physically ready ?
*/
static void _OnStateChange(unsigned IFaceId, U8 AdminState, U8 HWState) {
  //
  // Check if this is a disconnect from the peer or a link down.
  // In this case call IP_Disconnect() to get into a known state.
  //
  if (((AdminState == IP_ADMIN_STATE_DOWN) && (HWState == 1)) ||  // Typical for dial-up connection e.g. PPP when closed from peer. Link up but app. closed.
      ((AdminState == IP_ADMIN_STATE_UP)   && (HWState == 0))) {  // Typical for any Ethernet connection e.g. PPPoE. App. opened but link down.
    IP_Disconnect(IFaceId);                                       // Disconnect the interface to a clean state.
    MAIN_IPIsConnected = 0;
    BSP_ClrLED(IP_STATUS_LED);
  } else if ((AdminState == IP_ADMIN_STATE_UP)   && (HWState == 1)) {
    IP_Connect(IFaceId);
    MAIN_IPIsConnected = 1;
    BSP_SetLED(IP_STATUS_LED);
  }
}

/*********************************************************************
*
*       _Connect()
*
*  Function description
*    Open a socket.
*
*  Parameters
*    sHost - Host name to be resolved.
*    Port  - Port to open.
*
*  Return value
*    >= 0 - Success, socket handle.
*    <  0 - Error.
*/
static int _Connect(void *pVoid, const char *sHost, unsigned Port) {
  CONNECTION_CONTEXT* pConn;
  struct sockaddr_in  sa;
  int                 Status;
  U32                 IPAddr;

  pConn = pVoid;
  //
  // Decode host.
  //
  Status = IP_ResolveHost(sHost, &IPAddr, 1000);
  if (Status < 0) {
    return -1;
  }
  //
  pConn->Socket = socket(AF_INET, SOCK_STREAM, 0);
  if (pConn->Socket < 0) {
    Status = pConn->Socket;
  } else {
    sa.sin_family      = AF_INET;
    sa.sin_port        = htons(Port);
    sa.sin_addr.s_addr = IPAddr;
    Status = connect(pConn->Socket, (struct sockaddr *)&sa, sizeof(sa));
    if (Status < 0) {
      closesocket(pConn->Socket);
    }
  }
  return Status < 0 ? -1 : pConn->Socket;
}


/*********************************************************************
*
*       _Disconnect()
*
*  Function description
*    Disconnect from host.
*
*  Parameters
*    pVoid - Pointer to connection context.
*
*  Return value
*    >= 0 - Success.
*    <  0 - Processing error.
*/
static int _Disconnect(void *pVoid) {
  CONNECTION_CONTEXT* pConn;

  pConn = pVoid;
  if (pConn->Socket <= 0) {
    return 0;
  } else {
    return closesocket(pConn->Socket);
  }
}


/*********************************************************************
*
*       _Send()
*
*  Function description
*    Send data to socket.
*
*  Parameters
*    Socket  - Socket to write to.
*    pData   - Pointer to octet string to send.
*    DataLen - Octet length of the octet string to send.
*    Flags   - Socket send flags.
*
*  Return value
*    >= 0 - Number of bytes sent.
*    <  0 - Error.
*
*  Additional information
*    The number of bytes sent can be less than the number
*    of bytes that were requested to be written.
*/
static int _Send(void* pVoid, const void* pData, unsigned DataLen) {
  CONNECTION_CONTEXT* pConn;
  int                 Status;

  pConn  = pVoid;
  Status = send(pConn->Socket, pData, DataLen, 0);
  return Status < 0 ? -1 : Status;
}

/*********************************************************************
*
*       _Recv()
*
*  Function description
*    Receive data from socket.
*
*  Parameters
*    Socket  - Socket to read from.
*    pData   - Pointer to object that receives an octet string.
*    DataLen - Octet length of the octet string to receive.
*
*  Return value
*    >= 0 - Number of bytes received.
*    <  0 - Error.
*
*  Additional information
*    The number of bytes received can be less than the number
*    of bytes requested.
*/
static int _Recv(void* pVoid, void* pData, unsigned DataLen) {
  CONNECTION_CONTEXT* pConn;
  int                 Status;

  pConn  = pVoid;  
  Status = recv(pConn->Socket, pData, DataLen, 0);
  return Status < 0 ? -1 : Status;
}


/*********************************************************************
*
*       _PlainAPI
*/
static const IOT_IO_API _PlainAPI = {
  _Connect,
  _Disconnect,
  _Send,
  _Recv
};

/*********************************************************************
*
*       _GetWeather()
*
*  Function description
*    Sends a request to api.openweathermap.org and receives 
*    information about the current and future weather.
*/
static int _GetWeather(U32 GetNow, const char* host, const char* path) {
  IOT_HTTP_CONTEXT    IOT_HTTP_Con;
  IOT_HTTP_PARA       aPara[20];
  CONNECTION_CONTEXT  Connection;
  unsigned            StatusCode;
  int                 Status;
  int                 StatusHTTP;
  int                 StatusParse;
  IOT_JSON_CONTEXT    JSON;
  char                acBuf[32];
  USER_CONTEXT        User;

  User.pKey             = NULL;
  User.WeatherNow       = GetNow;
  User.WeatherDayIndex  = 0;
  IOT_JSON_Init(&JSON, &_EventAPI, &acBuf[0], sizeof(acBuf));
  //
  // Initialize the IOT Toolkit and send GET request
  //
  IOT_HTTP_Init      (&IOT_HTTP_Con, &_aBuf[0], sizeof(_aBuf), aPara, SEGGER_COUNTOF(aPara));
  IOT_HTTP_SetIO     (&IOT_HTTP_Con, &_PlainAPI, &Connection);
  IOT_HTTP_SetVersion(&IOT_HTTP_Con, &IOT_HTTP_VERSION_HTTP_1v1);
  IOT_HTTP_AddMethod (&IOT_HTTP_Con, "GET");
  IOT_HTTP_AddHost   (&IOT_HTTP_Con, host);
  IOT_HTTP_SetPort   (&IOT_HTTP_Con, 80);
  IOT_HTTP_AddPath   (&IOT_HTTP_Con, path);
  //
  Status = IOT_HTTP_Connect(&IOT_HTTP_Con);
  if (Status < 0) {
    IP_Logf_Application("Cannot negotiate a connection to %s:%d!", IOT_HTTP_GetHost(&IOT_HTTP_Con), IOT_HTTP_GetPort(&IOT_HTTP_Con));
    return Status;
  }
  //
  Status = IOT_HTTP_Exec(&IOT_HTTP_Con);
  if (Status < 0) {
    IP_Logf_Application("Cannot execute GET request!", IOT_HTTP_GetMethod(&IOT_HTTP_Con));
    goto Disconnect;
  }
  //
  Status = IOT_HTTP_ProcessStatusLine(&IOT_HTTP_Con, &StatusCode);
  if (Status < 0) {
    IP_Logf_Application("Cannot process status line!");
    goto Disconnect;
  }
  IP_Logf_Application("Returned status code: %u.", StatusCode);
  Status = IOT_HTTP_ProcessHeaders(&IOT_HTTP_Con, NULL);
  if (Status < 0) {
    IP_Logf_Application("Cannot process headers!");
    goto Disconnect;
  }
  IOT_HTTP_GetBegin(&IOT_HTTP_Con);
  while((StatusHTTP = IOT_HTTP_GetPayload(&IOT_HTTP_Con, &_aPayload[0], sizeof(_aPayload))) > 0) {
    IOT_JSON_SetUserContext(&JSON, &User);
    StatusParse = IOT_JSON_Parse(&JSON, _aPayload, StatusHTTP);
  }
  if (StatusParse == 0) {
    printf("Parse OK.\n");
    Status = 1;
  } else {
    printf("Parse FAIL\n");
    Status = 0;
  }
  IOT_HTTP_GetEnd(&IOT_HTTP_Con);
Disconnect:
  IOT_HTTP_Disconnect(&IOT_HTTP_Con);
  return Status;
}

/*********************************************************************
*
*       _SendGETRequest()
*
*  Function description
*   Sends a request to api.openweathermap.org, receives information 
*   about the current and future weather and parses it
*/
static int _SendGETRequest(void) {
  int i;
  int j;
  int p;
  int tMin;
  int tMax;

  j = 0;
  IP_Logf_Application("Weather now:");
  //
  // Get current weather status
  //
  if(!_GetWeather(1, WEATHER_API_SERVER, WEATHER_APP_WEATHER_NOW)) {
    return 0;
  }
  //
  // Append suffix
  //
  strcat(MAIN_WeatherNow.acTemp,    "C");
  strcat(MAIN_WeatherNow.acTempMin, "C");
  strcat(MAIN_WeatherNow.acTempMax, "C");
  strcat(MAIN_WeatherNow.acHumidity, "%");
  //
  // Get weather forecast
  //
  IP_Logf_Application("Weather forecast:");
  if(!_GetWeather(0, WEATHER_API_SERVER, WEATHER_APP_WEATHER_FORECAST)) { 
    return 0;
  }
  //
  tMax = atoll(_aWeatherDay[0].acTempMax); // Initialize tMax to the first value, so we won't get a conflict
  tMin = atoll(_aWeatherDay[0].acTempMin); // Initialize tMin to the first value, so we won't get a conflict
  //
  for(i = 0; i < 40; i++) {
    if(StringCompareEnd(_aWeatherDay[i].acDate, "00:00:00") && j < 5) { // We have a new day
      if(j > 0) {                                                       // Check if we are on a day we want to collect information about
        sprintf(MAIN_aWeatherDay[j-1].acTempMax, "%d", tMax);
        sprintf(MAIN_aWeatherDay[j-1].acTempMin, "%d", tMin);
      }
      tMax = atoll(_aWeatherDay[i].acTempMax);                          // Set tMax to the next value, so we won't get a conflict
      tMin = atoll(_aWeatherDay[i].acTempMin);                          // Set tMin to the next value, so we won't get a conflict
    }
    p = atoll(_aWeatherDay[i].acTempMax);
    if(p > tMax) {                                                      // Replace tMax if new value is larger
      tMax = p;
    }
    p = atoll(_aWeatherDay[i].acTempMin);
    if(p < tMin) {                                                      // Replace tMin if new value is smaller
      tMin = p;
    }
    if(StringCompareEnd(_aWeatherDay[i].acDate, "12:00:00") && j < 5) { // At 12:00 we collect information about weather status and humidity
      MAIN_aWeatherDay[j] = _aWeatherDay[i];
      strcat(MAIN_aWeatherDay[j].acHumidity, "%");                      // Append suffix
      p =  atoll(MAIN_aWeatherDay[j].acWeekday);
      strcpy(MAIN_aWeatherDay[j].acWeekday, GetWeekdayFromUTC(p));
      j++;
    }
  }
  sprintf(MAIN_aWeatherDay[j-1].acTempMax, "%d", tMax);                 // Set min and max temperature for the last day
  sprintf(MAIN_aWeatherDay[j-1].acTempMin, "%d", tMin);
  return  1;
}

/*********************************************************************
*
*       _TimerPulse
*/
static void _TimerPulse(void) {
  static int _TimeUpdateCount = 0;
  static int _SecondCount = 0;

  _SecondCount++;  
  if(_SecondCount % (1000/TIMER_UPDATE_PERIOD) == 0) { // One secons has passed
    _SecondCount = 0;
    GLOBAL_Time.Seconds++;
    _TimeUpdateCount++;
    //
    if (MAIN_IPIsConnected) {
      if((_TimeUpdateCount > TIME_UPDATE_PERIOD) || !MAIN_TimeIsInitialized) {
        _TimeUpdateCount     = 0;
        MAIN_TimeNeedsUpdate = 1;
      }
    }
  }
  OS_TIMER_Restart(&_TIMERPulse); 
}

/*********************************************************************
*
*       _InitWeatherApp()
*
* Function description
*   Initializes the application settings.
*/
void _InitWeatherApp(void) {
  int i;

  // 
  // Initialize weather
  //
  strcpy(MAIN_WeatherNow.acTemp,    "-----C");
  strcpy(MAIN_WeatherNow.acTempMin, "-----C");
  strcpy(MAIN_WeatherNow.acTempMax, "-----C");
  strcpy(MAIN_WeatherNow.acWeather, "-----");
  //
  for(i = 0; i < 5; i++) {
    strcpy(MAIN_aWeatherDay[i].acWeekday,    "---");
    strcpy(MAIN_aWeatherDay[i].acTempMax,    "---");
    strcpy(MAIN_aWeatherDay[i].acTempMin,    "---");
    strcpy(MAIN_aWeatherDay[i].acTemp,       "---");
    strcpy(MAIN_aWeatherDay[i].acHumidity,   "--%");
    strcpy(MAIN_aWeatherDay[i].acWeather,    "---");
  }
  //
  MAIN_TempNeedsUpdate   = 1;
  MAIN_TimeNeedsUpdate   = 1;
  MAIN_TimeIsInitialized = 0;
}

/*********************************************************************
*
*       Public code
*
**********************************************************************
*/

/*********************************************************************
*
*       MainTask_HTTP()
*
* Function description
*   Task executed by the RTOS to create further resources and
*   running the main application.
*/
void MainTask_HTTP(void) {
  //
  // Initialize the application default settings.
  //
  _InitWeatherApp();
  //
  // Create software timer for accurate timing
  //
  OS_TIMER_CREATE(&_TIMERPulse, _TimerPulse, 100);   
  OS_TIMER_Start(&_TIMERPulse);                      // Start the timer
  //
  // Initialize embOS/IP
  //
  IP_Init();
  OS_CREATETASK(&_IPTCB  , "IP_Task"  , IP_Task   , TASK_PRIO_IP_TASK   , _IPStack);    // Start the IP task
#if USE_RX_TASK 
  OS_CREATETASK(&_IPRxTCB, "IP_RxTask", IP_RxTask , TASK_PRIO_IP_RX_TASK, _IPRxStack);  // Start the IP_RxTask, optional.
#endif
  IP_AddStateChangeHook(&_StateChangeHook, _OnStateChange);                            // Register hook to be notified on disconnects.
  IP_Connect(_IFaceId);                                                                // Connect the interface if necessary.
  //
  while (IP_IFaceIsReadyEx(_IFaceId) == 0) {
    BSP_ToggleLED(1);
    OS_Delay(50);
  }
  //
  while (1) {
    BSP_ToggleLED(0);
    if (MAIN_IPIsConnected && _SendGETRequest() >= 0) { // Get the current weather information
      MAIN_TempNeedsUpdate = 1;                         // If no error occurred, we need to update the display
      OS_Delay(WEATHER_UPDATE_PERIOD);                  // Delay task to only update weather after a specific delay
    } else {
      OS_Delay(WEATHER_FAIIL_PERIOD);                   // An error occurred, the task will be delayed to try again
    }
  }
}

/*************************** End of file ****************************/
