/*********************************************************************
*                      SEGGER Microcontroller GmbH                   *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*      (c) 2018  SEGGER Microcontroller GmbH                         *
*                                                                    *
*       www.segger.com     Support: support@segger.com               *
*                                                                    *
**********************************************************************
-------------------------- END-OF-HEADER -----------------------------
File    : UTIL.c
Purpose : Helper functions.
*/

/*********************************************************************
*
*       #include Section
*
**********************************************************************
*/

#include "WeatherStation.h"

/*********************************************************************
*
*       Defines
*
**********************************************************************
*/

#define SECONDS_PER_DAY       86400L
#define SECONDS_PER_70YEARS   2208988800UL

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/

static U8 _aDaysInMonth[] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};

/*********************************************************************
*
*       Static const data
*
**********************************************************************
*/

static const char* const _WeekDays[] = {
  "Thu",
  "Fri",
  "Sat",
  "Sun",
  "Mon",
  "Tue",
  "Wed"
};


/*********************************************************************
*
*       Static code
*
**********************************************************************
*/

/*********************************************************************
*
*       UTIL_GetTimeDate()
*
*  Function description
*    Returns time and date of the system.
*
*  Return value
*    Date and time of the system as U32.
*/
U32 UTIL_GetTimeDate(void) {
  UTIL_DATE_TIME DateTime;
  U32            r;
  U16            Sec;
  U16            Min;
  U16            Hour;
  U16            Day;
  U16            Month;
  U16            Year;

  r = 0;
  DateTime = UTIL_GetTimeDateFromTimestamp(GLOBAL_Time.Seconds, 1900);
  Sec      = DateTime.Second;       //    0 based. Valid range: 0..59 .
  Min      = DateTime.Minute;       //    0 based. Valid range: 0..59 .
  Hour     = DateTime.Hour;         //    0 based. Valid range: 0..23 .
  Day      = DateTime.Day;          //    1 based. Means that 1 is 1. Valid range is 1..31 (depending on month).
  Month    = DateTime.Month;        //    1 based. Means that January is 1. Valid range is 1..12 .
  Year     = DateTime.Year - 1980;  // 1980 based. Means that 2008 would be 28.
  r        = (Sec / 2) + (Min << 5) + (Hour  << 11);
  r       |= ((U32)(Day + (Month << 5) + (Year  << 9))) << 16;
  return r;
}


/*********************************************************************
*
*       UTIL_GetTimeDateFromTimestamp()
*
*  Function description
*    Returns time and date of the system.
*
*  Return value
*    Date and time of the system.
*/
UTIL_DATE_TIME UTIL_GetTimeDateFromTimestamp(U32 Timestamp, U32 Epoche) {
  UTIL_DATE_TIME DateTime;
  U32            Day;
  U8*            pMonth;
  U8             IsLeapYear;

  Day             = Timestamp / (3600 * 24);
  DateTime.Hour   = (Timestamp / 3600) - (Day * 24);
  DateTime.Minute = (Timestamp / 60) - (Day * 24 * 60) - (DateTime.Hour * 60);
  DateTime.Second = Timestamp - (Day * 24 * 3600) - (DateTime.Hour * 3600) - (DateTime.Minute * 60);
  //
  // Get year
  //
  DateTime.Year = Epoche;
  if ((DateTime.Year % 4 == 0) && ((DateTime.Year % 100 != 0) || (DateTime.Year % 400 == 0))) {  // Check if leap year
    IsLeapYear = 1;
  } else {
    IsLeapYear = 0;
  }
  while (Day > (365u + IsLeapYear)) {
    DateTime.Year++;
    Day -= (365 + IsLeapYear);
    if ((DateTime.Year % 4 == 0) && ((DateTime.Year % 100 != 0) || (DateTime.Year % 400 == 0))) {  // Check if leap year
      IsLeapYear = 1;
    } else {
      IsLeapYear = 0;
    }
  }
  //
  // Get month
  //
  DateTime.Month = 1;
  if (IsLeapYear) {
    _aDaysInMonth[1] = 29;
  } else {
    _aDaysInMonth[1] = 28;
  }
  pMonth = &_aDaysInMonth[0];
  while (Day >= *pMonth) {
    Day -= *pMonth;
    pMonth++;
    DateTime.Month++;
  }
  // Month and Day at starting at 1 and not 0
  DateTime.Day = Day + 1;
  //
  return DateTime;
}


/*********************************************************************
*
*       UTIL_PrintDateTime()
*
*  Function description
*    TODO
*
*  Return value
*    TODO
*/
int UTIL_PrintDateTime(char* pBuffer, int NumBytesBuffer, U32 Timestamp, U32 Epoche) {
  UTIL_DATE_TIME DateTime;

  DateTime = UTIL_GetTimeDateFromTimestamp(Timestamp, Epoche);
  return SEGGER_snprintf(pBuffer, NumBytesBuffer, "%d.%d.%d %d:%d:%d", DateTime.Day, DateTime.Month, DateTime.Year, DateTime.Hour, DateTime.Minute, DateTime.Second);
}


/*********************************************************************
*
*       UTIL_IPGetNTPTime()
*
* Function description
*   Get the current time from an NTP server.
*
* Parameters
*   IFaceId   : Zero-based interface index.
*/
int UTIL_IPGetNTPTime(int IFaceId, int TimezoneOffset) {
  IP_NTP_TIMESTAMP Time;
  int              Status;

  //
  // Check if interface is ready.
  //
  while (IP_IFaceIsReadyEx(IFaceId) == 0) {
    return 0;
  }
  //
  // Get the current time from an NTP server
  //
  IP_Logf_Application("Requesting time stamp from %s:", NTP_SERVER);
  Status = IP_SNTPC_GetTimeStampFromServer(IFaceId, NTP_SERVER, &Time);
  if (Status < 0) {
    IP_Logf_Application("Communication error.\n");
  } else {
    switch (Status) {
    case IP_SNTPC_STATE_NO_ANSWER:
      IP_Logf_Application("Timeout.");
      break;
    case IP_SNTPC_STATE_UPDATED:
      IP_Logf_Application("Timestamp received: %lu seconds passed since January 1st 1900.", Time.Seconds);
      GLOBAL_Time = Time;
      GLOBAL_Time.Seconds += TimezoneOffset;
      return 1;
      break;
    case IP_SNTPC_STATE_KOD:
      IP_Logf_Application("Kiss-Of-Death received. Use another server.");
      break;
    default:
      break;
    }
  }
  return 0;
}


/*********************************************************************
*
*       GetWeekdayFromUTC()
*
* Function description
*   Returns the weekday based on a UTC-Timestamp
*
* Parameters
*   Time   : UTC-Timestamp
*/
const char* GetWeekdayFromUTC(U32 Time) {
  long Day = Time / SECONDS_PER_DAY;

  Day = Time / SECONDS_PER_DAY;
  return _WeekDays[(Day % 7)];
}


/*********************************************************************
*
*       StringCompareEnd()
*
* Function description
*   Returns 1 if the end of string a is equal to b
*
* Parameters
*   a   : String to check
*   b   : The string to be searched for at the end of string a
*/
int StringCompareEnd(const char* s, const char* sMatch) {
  int Diff;

  Diff = strlen(s) - strlen(sMatch);
  if(Diff < 0) {
    return 0;
  }
  if(strcmp(&s[Diff], sMatch) == 0) {
    return 1;
  }
  return 0;
}


/*********************************************************************
*
*       GetWeatherIconId()
*
* Function description
*   Returns the weather-icon id
*
* Parameters
*   sIcon: String for icon
*/
U32 GetWeatherIconId(const char* sIcon) {
  U32 v;

  if (strncmp(sIcon, "01", 2) == 0) {
    v = WEATHER_ICON_CLEAR_SKY;
  } else if (strncmp(sIcon, "02", 2) == 0) {
    v = WEATHER_ICON_FEW_CLOUDS;
  } else if (strncmp(sIcon, "03", 2) == 0) {
    v = WEATHER_ICON_SCATTERED_CLOUDS;
  } else if (strncmp(sIcon, "04", 2) == 0) {
    v = WEATHER_ICON_BROKEN_CLOUDS;
  } else if (strncmp(sIcon, "09", 2) == 0) {
    v = WEATHER_ICON_SHOWER_RAIN;
  } else if (strncmp(sIcon, "10", 2) == 0) {
    v = WEATHER_ICON_RAIN;
  } else if (strncmp(sIcon, "11", 2) == 0) {
    v = WEATHER_ICON_THUNDERSTORM;
  } else if (strncmp(sIcon, "13", 2) == 0) {
    v = WEATHER_ICON_SNOW;
  } else if (strncmp(sIcon, "50", 2) == 0) {
    v = WEATHER_ICON_MIST;
  } else {
    v = WEATHER_ICON_UNKNOWN;
  }
  return v;
}

/****** End Of File *************************************************/